﻿using Microsoft.Extensions.Configuration;
using Phares.Shared;
using Serilog;
using ShellProgressBar;
using System.Drawing;
using System.Drawing.Imaging;
using System.Reflection;
using System.Text;
using View_by_Distance.Offset.Date.Time.Original.Models;
using View_by_Distance.Shared.Models;
using View_by_Distance.Shared.Models.Methods;
using View_by_Distance.Shared.Models.Stateless;

namespace View_by_Distance.Offset.Date.Time.Original;

public class OffsetDateTimeOriginal
{

    private record Record(FileHolder FileHolder,
                          bool IsIgnoreExtension,
                          bool IsValidImageFormatExtension,
                          int Id,
                          DateTime DateTime);

    private readonly AppSettings _AppSettings;
    private readonly string _WorkingDirectory;
    private readonly Configuration _Configuration;
    private readonly IsEnvironment _IsEnvironment;
    private readonly IConfigurationRoot _ConfigurationRoot;
    private readonly Property.Models.Configuration _PropertyConfiguration;

    public OffsetDateTimeOriginal(List<string> args, IsEnvironment isEnvironment, IConfigurationRoot configurationRoot, AppSettings appSettings, string workingDirectory, bool isSilent, IConsole console)
    {
        if (isSilent)
        { }
        if (console is null)
        { }
        _AppSettings = appSettings;
        _IsEnvironment = isEnvironment;
        _WorkingDirectory = workingDirectory;
        _ConfigurationRoot = configurationRoot;
        ILogger? log = Log.ForContext<OffsetDateTimeOriginal>();
        Property.Models.Configuration propertyConfiguration = Property.Models.Binder.Configuration.Get(isEnvironment, configurationRoot);
        Configuration configuration = Models.Binder.Configuration.Get(isEnvironment, configurationRoot, propertyConfiguration);
        _PropertyConfiguration = propertyConfiguration;
        _Configuration = configuration;
        propertyConfiguration.Update();
        log.Information(propertyConfiguration.RootDirectory);
        Verify();
        List<string> lines = OffsetDateTimeOriginalFilesInDirectories(log);
        if (!lines.Any())
            File.WriteAllLines($"D:/Tmp/Phares/{DateTime.Now.Ticks}.tsv", lines);
    }

    private void Verify()
    {
        if (_AppSettings is null)
        { }
        if (_IsEnvironment is null)
        { }
        if (_Configuration is null)
        { }
        if (_ConfigurationRoot is null)
        { }
        if (_WorkingDirectory is null)
        { }
        if (_PropertyConfiguration is null)
        { }
    }

    private List<Record> GetRecords(ASCIIEncoding asciiEncoding, string checkDirectory, DateTime minimumDateTime, DateTime maximumDateTime, long ticks)
    {
        List<Record> results = new();
        DateTime dateTime;
        int? id;
        string? message;
        DateTime[] dateTimes;
        FileHolder fileHolder;
        bool isIgnoreExtension;
        DateTime? dateTimeOriginal;
        bool isValidImageFormatExtension;
        string[] files = Directory.GetFiles(checkDirectory, "*", SearchOption.TopDirectoryOnly);
        foreach (string file in files)
        {
            fileHolder = new(file);
            isValidImageFormatExtension = _PropertyConfiguration.ValidImageFormatExtensions.Contains(fileHolder.ExtensionLowered);
            isIgnoreExtension = isValidImageFormatExtension && _PropertyConfiguration.IgnoreExtensions.Contains(fileHolder.ExtensionLowered);
            (dateTimeOriginal, dateTimes, id, message) = Property.Models.Stateless.IProperty.Get(_PropertyConfiguration.PopulatePropertyId, fileHolder, isIgnoreExtension, isValidImageFormatExtension, asciiEncoding);
            if (message is not null)
                throw new Exception(message);
            if (id is null || dateTimeOriginal is null)
                continue;
            dateTime = dateTimeOriginal.Value.AddTicks(ticks);
            if (dateTime < minimumDateTime)
                continue;
            if (dateTime > maximumDateTime)
                continue;
            results.Add(new(fileHolder, isIgnoreExtension, isValidImageFormatExtension, id.Value, dateTimeOriginal.Value.AddTicks(ticks)));
        }
        if (files.Length != results.Count)
            throw new Exception();
        return results;
    }

#pragma warning disable CA1416

    private void DateFix(string sourceDirectory, ASCIIEncoding asciiEncoding, string checkDirectory, DateTime minimumDateTime, DateTime maximumDateTime, long ticks)
    {
        int? id;
        Bitmap bitmap;
        short type = 2;
        string checkFile;
        PropertyItem? propertyItem;
        string? ticksDirectory = null;
        int dateTimeOriginal = (int)IExif.Tags.DateTimeOriginal;
        for (int i = 0; i < int.MaxValue; i++)
        {
            ticksDirectory = Path.Combine(sourceDirectory, ticks.ToString());
            if (!Directory.Exists(ticksDirectory))
            {
                _ = Directory.CreateDirectory(ticksDirectory);
                break;
            }
            ticks++;
        }
        List<Record> records = GetRecords(asciiEncoding, checkDirectory, minimumDateTime, maximumDateTime, ticks);
        ConstructorInfo? constructorInfo = typeof(PropertyItem).GetConstructor(BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.Public, null, Array.Empty<Type>(), null) ?? throw new Exception();
        string? message = nameof(OffsetDateTimeOriginal);
        ProgressBarOptions options = new() { ProgressCharacter = '─', ProgressBarOnBottom = true, DisableBottomPercentage = true };
        ProgressBar progressBar = new(records.Count, message, options);
        foreach (Record record in records)
        {
            progressBar.Tick();
            if (ticksDirectory is null)
                throw new Exception();
            checkFile = Path.Combine(ticksDirectory, record.FileHolder.Name);
            if (File.Exists(checkFile))
                continue;
            propertyItem = Property.Models.Stateless.IProperty.GetPropertyItem(constructorInfo, dateTimeOriginal, type, record.DateTime.ToString("yyyy:MM:dd HH:mm:ss"));
            bitmap = new(record.FileHolder.FullName);
            bitmap.SetPropertyItem(propertyItem);
            bitmap.Save(checkFile);
            bitmap.Dispose();
            (_, _, id, message) = Property.Models.Stateless.IProperty.Get(_PropertyConfiguration.PopulatePropertyId, record.FileHolder, record.IsIgnoreExtension, record.IsValidImageFormatExtension, asciiEncoding);
            if (message is not null)
                throw new Exception(message);
            if (id is null || id.Value != record.Id)
                throw new Exception();
        }
        progressBar.Dispose();
    }

#pragma warning restore CA1416

    private void DateFix(ILogger log, string sourceDirectory)
    {
        string checkDirectory;
        ASCIIEncoding asciiEncoding = new();
        long oneYearTicks = DateTime.MinValue.AddYears(1).Ticks;
        checkDirectory = Path.Combine(sourceDirectory, oneYearTicks.ToString());
        if (Directory.Exists(checkDirectory))
            DateFix(sourceDirectory, asciiEncoding, checkDirectory, DateTime.MinValue, DateTime.MaxValue, new TimeSpan(oneYearTicks - DateTime.MinValue.Ticks).Ticks);
        else
        {
            checkDirectory = Path.Combine(sourceDirectory, "1");
            if (!Directory.Exists(checkDirectory))
                log.Error($"<{checkDirectory}> doesn't exist!");
            else
            {
                string badDirectory = Path.Combine(sourceDirectory, "Bad");
                string targetDirectory = Path.Combine(sourceDirectory, "Target");
                string[] minimumDirectory = Directory.GetDirectories(targetDirectory, "*", SearchOption.TopDirectoryOnly);
                if (minimumDirectory.Length != 1)
                    log.Error($"<{checkDirectory}> doesn't exist!");
                else
                {
                    string format = "yyyy-MM-dd";
                    string[] maximumDirectory = Directory.GetDirectories(minimumDirectory.First(), "*", SearchOption.TopDirectoryOnly);
                    string[] badFiles = !Directory.Exists(badDirectory) ? Array.Empty<string>() : Directory.GetFiles(badDirectory, "*", SearchOption.TopDirectoryOnly);
                    string[] targetFiles = !Directory.Exists(targetDirectory) ? Array.Empty<string>() : Directory.GetFiles(targetDirectory, "*", SearchOption.TopDirectoryOnly);
                    if (badFiles.Length != 1 || targetFiles.Length != 1 || maximumDirectory.Length != 1)
                        log.Error("bad file(s) or target file(s) or maximum directory doesn't equal 1!");
                    else
                    {
                        int? badId;
                        int? targetId;
                        string? badMessage;
                        string? targetMessage;
                        DateTime[] badDateTimes;
                        bool badIsIgnoreExtension;
                        DateTime[] targetDateTimes;
                        bool targetIsIgnoreExtension;
                        DateTime? badDateTimeOriginal;
                        DateTime? targetDateTimeOriginal;
                        FileHolder badFileHolder = new(badFiles.First());
                        FileHolder targetFileHolder = new(targetFiles.First());
                        bool badIsValidImageFormatExtension = _PropertyConfiguration.ValidImageFormatExtensions.Contains(badFileHolder.ExtensionLowered);
                        bool targetIsValidImageFormatExtension = _PropertyConfiguration.ValidImageFormatExtensions.Contains(targetFileHolder.ExtensionLowered);
                        badIsIgnoreExtension = badIsValidImageFormatExtension && _PropertyConfiguration.IgnoreExtensions.Contains(badFileHolder.ExtensionLowered);
                        targetIsIgnoreExtension = targetIsValidImageFormatExtension && _PropertyConfiguration.IgnoreExtensions.Contains(targetFileHolder.ExtensionLowered);
                        DateTime minimumDateTime = DateTime.ParseExact(Path.GetFileName(minimumDirectory.First()), format, null, System.Globalization.DateTimeStyles.None);
                        DateTime maximumDateTime = DateTime.ParseExact(Path.GetFileName(maximumDirectory.First()), format, null, System.Globalization.DateTimeStyles.None).AddHours(23);
                        (badDateTimeOriginal, badDateTimes, badId, badMessage) = Property.Models.Stateless.IProperty.Get(_PropertyConfiguration.PopulatePropertyId, badFileHolder, badIsIgnoreExtension, badIsValidImageFormatExtension, asciiEncoding);
                        if (badMessage is not null)
                            throw new Exception(badMessage);
                        if (!badDateTimes.Any() || badId is null)
                            throw new Exception(badMessage);
                        (targetDateTimeOriginal, targetDateTimes, targetId, targetMessage) = Property.Models.Stateless.IProperty.Get(_PropertyConfiguration.PopulatePropertyId, targetFileHolder, targetIsIgnoreExtension, targetIsValidImageFormatExtension, asciiEncoding);
                        if (targetMessage is not null)
                            throw new Exception(targetMessage);
                        if (!targetDateTimes.Any() || targetId is null)
                            throw new Exception(targetMessage);
                        if (badDateTimeOriginal is null || targetDateTimeOriginal is null)
                            log.Error("Date is null!");
                        else
                        {
                            TimeSpan timeSpan = new(targetDateTimeOriginal.Value.Ticks - badDateTimeOriginal.Value.Ticks);
                            DateFix(sourceDirectory, asciiEncoding, checkDirectory, minimumDateTime, maximumDateTime, timeSpan.Ticks);
                        }
                    }
                }
            }
        }
    }

    private List<string> OffsetDateTimeOriginalFilesInDirectories(ILogger log)
    {
        List<string> results = new();
        DateFix(log, _PropertyConfiguration.RootDirectory);
        return results;
    }

}